"""Import and parse the data files produced by the Marianda VINDTA 3C."""

import re
import numpy as np, pandas as pd
from matplotlib import dates as mdates


def add_func_cols(df, func, *args, **kwargs):
    """Add results of df.apply(func) to df as new columns."""
    return df.join(df.apply(lambda x: func(x, *args, **kwargs), axis=1))


def dbs_datetime(dbsx):
    """Convert date and time from .dbs file into datetime and datenum."""
    dspl = dbsx["date"].split("/")
    analysis_datetime = np.datetime64(
        "-".join(("20" + dspl[2], dspl[0], dspl[1])) + "T" + dbsx["time"]
    )
    return pd.Series(
        {
            "analysis_datetime": analysis_datetime,
            "analysis_datenum": mdates.date2num(analysis_datetime),
        }
    )


def read_dbs(fname):
    """Import one .dbs file as single DataFrame."""
    headers = np.genfromtxt(fname, delimiter="\t", dtype=str, max_rows=1)
    dbs = pd.read_table(fname, header=0, names=headers, usecols=headers)
    dbs["dbs_fname"] = fname
    dbs = add_func_cols(dbs, dbs_datetime)
    return dbs


def read_logfile(fname, methods="3C standard"):
    """Import a logfile.bak as a DataFrame."""
    if isinstance(methods, str):
        methods = [methods]
    # Compile regexs for reading logfile
    re_method = re.compile(
        r"(" + r"|".join(methods) + r")\.mth run started ".format(methods)
    )
    re_datetime = re.compile(r"started (\d{2})/(\d{2})/(\d{2})  (\d{2}):(\d{2})")
    re_bottle = re.compile(r"(bottle)?\t([^\t]*)\t")
    re_crm = re.compile(r"CRM\t([^\t]*)\t")
    re_increments = re.compile(r"(\d*)\t(\d*)\t(\d*)\t")
    # Import text from logfile
    with open(fname, "r") as f:
        logf = f.read().splitlines()
    # Initialise arrays
    logdf = {
        "logfile_line": [],
        "analysis_datetime": np.array([], dtype="datetime64"),
        "bottle": [],
        "table": [],
        "counts_total": [],
        "runtime": [],
        "method": [],
    }
    # Parse file line by line
    for i, line in enumerate(logf):
        if re_method.match(line):
            logdf["logfile_line"].append(i)
            logdf["method"].append(re_method.findall(line)[0])
            # Get analysis date and time
            ldt = re_datetime.findall(line)[0]
            ldt = np.datetime64(
                "{}-{}-{}T{}:{}".format("20" + ldt[2], ldt[0], ldt[1], ldt[3], ldt[4])
            )
            logdf["analysis_datetime"] = np.append(logdf["analysis_datetime"], ldt)
            # Get sample name
            lbot = 0
            if re_bottle.match(logf[i + 1]):
                lbot = re_bottle.findall(logf[i + 1])[0][1]
            elif re_crm.match(logf[i + 1]):
                lbot = re_crm.findall(logf[i + 1])[0]
            elif logf[i + 1] == "other":
                lbot = "other_{}".format(i + 1)
            assert type(lbot) == str, "Logfile line {}: bottle name not found!".format(
                i + 1
            )
            logdf["bottle"].append(lbot)
            # Get coulometer data
            jdict = {"minutes": [0.0], "counts": [0.0], "increments": [0.0]}
            j = 4
            while re_increments.match(logf[i + j].strip()):
                jinc = re_increments.findall(logf[i + j].strip())[0]
                jdict["minutes"].append(float(jinc[0]))
                jdict["counts"].append(float(jinc[1]))
                jdict["increments"].append(float(jinc[2]))
                j += 1
            jdict = {k: np.array(v) for k, v in jdict.items()}
            logdf["table"].append(jdict)
            logdf["counts_total"].append(jdict["counts"][-1])
            logdf["runtime"].append(j - 4.0)
    # Convert lists to arrays and put logfile into DataFrame
    logdf = {k: np.array(v) for k, v in logdf.items()}
    return pd.DataFrame(logdf)


def get_logfile_index(dbs_row, logfile):
    """Get index in logfile corresponding to a given row of the dbs file."""
    if dbs_row.bottle in logfile.bottle.values:
        logfile_index = np.where(
            (dbs_row.bottle == logfile.bottle)
            & (dbs_row.analysis_datetime == logfile.analysis_datetime)
        )[0]
        assert np.size(logfile_index) == 1, (
            "{} name/date matches found between dbs and logfile @ dbs loc {}"
        ).format(np.size(logfile_index), dbs_row.name)
        logfile_index = logfile.index[logfile_index[0]]
    else:
        logfile_index = np.nan
    return logfile_index


def logfile2dbs(dbs_row, logfile):

    if dbs_row.bottle in logfile.bottle.values:
        xix = np.where(
            (dbs_row.bottle == logfile.bottle)
            & (dbs_row.analysis_datetime == logfile.analysis_datetime)
        )[0]
        assert len(xix) == 1, (
            "{} name/date matches found between dbs and logfile @ dbs loc {}"
        ).format(len(xix), dbs_row.name)
        xix = logfile.index[xix[0]]
    else:
        xix = np.nan
    return pd.Series({"logfile_index": xix})


# def logfile2dbs(dbs, logfile):
#     """Get index in logfile corresponding to each row in dbs."""

#     def _logfile2dbs(x, logfile):
#         if x.bottle in logfile.bottle.values:
#             xix = np.where(
#                 (x.bottle == logfile.bottle)
#                 & (x.analysis_datetime == logfile.analysis_datetime)
#             )[0]
#             assert len(xix) == 1, (
#                 "{} name/date matches found between dbs and logfile @ dbs iloc {}"
#             ).format(len(xix), x.name)
#             xix = logfile.index[xix[0]]
#         else:
#             xix = np.nan
#         return pd.Series({"logfile_index": xix})

#     return add_func_cols(dbs, _logfile2dbs, logfile)



"""Import and parse the data files produced by the Marianda VINDTA 3C."""

from .dbs import Dbs, concat
from .read import read_dbs, read_logfile
from .get import (
    get_logfile_index,
    get_sample_blanks,
    blank_progression,
    centre_and_scale,
    de_centre_and_scale,
    get_session_blanks,
    blank_correction,
    get_blank_corrections,
    get_density,
    get_standard_calibrations,
    get_session_calibrations,
    calibrate_dic,
)
from .plot import (
    plot_session_blanks,
    plot_blanks,
    plot_k_dic,
    plot_dic_offset,
)
from .process import poison_correction
from . import io, plot, process



import numpy as np
import pandas as pd
from scipy.optimize import least_squares
from . import io


def get_sample_blanks(dbs_row, logfile, use_from=6):
    lft = logfile.loc[dbs_row.logfile_index].table
    use_minutes = lft["minutes"] >= use_from
    blank_here = np.mean(lft["increments"][use_minutes])
    return blank_here


def centre_and_scale(x, x_factor=None, x_offset=None):
    if x_factor is None:
        x_factor = np.std(x)
    if x_offset is None:
        x_offset = np.mean(x)
    return (x - x_offset) / x_factor


def de_centre_and_scale(x, x_factor, x_offset):
    return x * x_factor + x_offset


def blank_progression(x0, datenum_scaled):
    """Estimate the changing coulometer blank during an analysis session."""
    return (
        x0[0]
        + x0[1] * datenum_scaled
        + x0[2] * np.exp(-(datenum_scaled - x0[3]) / x0[4])
    )


def fit_blank_progression(x0, datenum_scaled, blank_here):
    """Fit the changing coulometer blank during an analysis session."""
    return blank_progression(x0, datenum_scaled) - blank_here


def get_session_blanks(x):
    """Calculate blanks per analysis session."""
    blank_here = x[x.blank_good].blank_here
    datenum_here = x[x.blank_good].analysis_datenum
    datenum_mean = np.mean(datenum_here)
    datenum_std = np.std(datenum_here)
    datenum_scaled = centre_and_scale(
        datenum_here, x_factor=datenum_std, x_offset=datenum_mean
    )
    # blank_polyfit = np.polyfit(datenum_scaled, blank_here, 2)
    blank_prog = least_squares(
        fit_blank_progression, [0, 1, 1, 1, 1], args=[datenum_scaled, blank_here]
    )
    return pd.Series(
        data={
            "blank_mean": blank_here.mean(),
            "blank_median": blank_here.median(),
            "analysis_datenum_mean": datenum_mean,
            "analysis_datenum_std": datenum_std,
            # "blank_polyfit": blank_polyfit,
            "blank_progression": blank_prog["x"],
        }
    )


def blank_correction(
    dbs, blank_col="blank", counts_col="counts", runtime_col="run time"
):
    """Apply the blank correction to the coulometer counts."""
    return dbs[counts_col] - dbs[runtime_col] * dbs[blank_col]


def apply_session_blanks(dbs, dic_sessions):
    """Estimate sample blanks from session fits and apply the blank correction."""
    dbs["analysis_datenum_scaled"] = np.nan
    dbs["blank"] = np.nan
    for session in dic_sessions.index:
        S = dbs["cell ID"] == session
        dbs.loc[S, "analysis_datenum_scaled"] = centre_and_scale(
            dbs.loc[S].analysis_datenum,
            x_factor=dic_sessions.loc[session].analysis_datenum_std,
            x_offset=dic_sessions.loc[session].analysis_datenum_mean,
        )
        dbs.loc[S, "blank"] = blank_progression(
            dic_sessions.loc[session].blank_progression,
            dbs.loc[S].analysis_datenum_scaled,
        )
    dbs["counts_corrected"] = blank_correction(dbs)
    return dbs


def get_data_index(dbs_row, data):
    """Get index of data in dbs to transfer metadata across for processing."""
    if dbs_row.bottle in data.station_bottleid.values:
        data_iloc = np.where(dbs_row.bottle == data.station_bottleid.values)[0]
        assert (
            len(data_iloc) == 1
        ), "Found more than one bottle ID match for {}!".format(dbs_row.bottle)
        data_index = data.index[data_iloc[0]]
    else:
        data_index = np.nan
    return data_index


def dbs2data(dbs, data, fields):
    """Transfer results from dbs into data, taking means for duplicates."""
    if isinstance(fields, str):
        fields = [fields]
    for field in fields:
        data[field] = np.nan
        data[field + "_std"] = np.nan
        data[field + "_count"] = 0
    for row in data.index:
        dr = data.loc[row]
        if dr.station_bottleid in dbs.bottle.values:
            for field in fields:
                dbs_iloc = np.where(
                    (dr.station_bottleid == dbs.bottle.values) & dbs[field + "_good"]
                )[0]
                if np.size(dbs_iloc) > 0:
                    fdata = dbs.iloc[dbs_iloc][field]
                    data.loc[row, field] = np.mean(fdata)
                    if np.size(dbs_iloc) > 1:
                        data.loc[row, field + "_std"] = np.std(fdata)
                    data.loc[row, field + "_count"] = np.size(dbs_iloc)
    return data


def poison_correction(var, sample_volume, poison_volume):
    """Apply dilution correction to var for added poison."""
    return var * (1 + poison_volume / sample_volume)



import numpy as np, pandas as pd
from scipy.optimize import least_squares
from calkulate import density


def _get_logfile_index(dbs_row, logfile):
    """[row.apply] Get index in logfile corresponding to a given row of the dbs file."""
    if dbs_row.bottle in logfile.bottle.values:
        logfile_index = np.where(
            (dbs_row.bottle == logfile.bottle)
            & (dbs_row.datetime_analysis == logfile.datetime_analysis)
        )[0]
        assert np.size(logfile_index) == 1, (
            "{} name/date matches found between dbs and logfile @ dbs loc {}"
        ).format(np.size(logfile_index), dbs_row.name)
        logfile_index = logfile.index[logfile_index[0]]
    else:
        logfile_index = np.nan
    return logfile_index


def get_logfile_index(dbs):
    """Get index in logfile corresponding to all rows of the dbs file."""
    assert dbs.logfile is not None, "You have not assigned a logfile for this dbs!"
    dbs["logfile_index"] = dbs.apply(_get_logfile_index, args=[dbs.logfile], axis=1)
    return dbs


def _get_sample_blanks(dbs_row, logfile, use_from=6):
    """[row.apply] Calculate each sample's DIC blank value."""
    lft = logfile.loc[dbs_row.logfile_index].table
    use_minutes = lft["minutes"] >= use_from
    blank_here = np.mean(lft["increments"][use_minutes])
    return blank_here


def get_sample_blanks(dbs, use_from=6):
    """Calculate each sample's DIC blank value."""
    if "logfile_index" not in dbs:
        dbs.get_logfile_index()
    dbs["blank_here"] = dbs.apply(
        _get_sample_blanks, args=[dbs.logfile], axis=1, use_from=use_from
    )
    return dbs


def centre_and_scale(x, x_factor=None, x_offset=None):
    if x_factor is None:
        x_factor = np.std(x)
    if x_offset is None:
        x_offset = np.mean(x)
    return (x - x_offset) / x_factor


def de_centre_and_scale(x, x_factor, x_offset):
    return x * x_factor + x_offset


def blank_progression(x0, datenum_scaled):
    """Estimate the changing coulometer blank during an analysis session."""
    return (
        x0[0]
        + x0[1] * datenum_scaled
        + x0[2] * np.exp(-(datenum_scaled - x0[3]) / x0[4])
    )


def fit_blank_progression(x0, datenum_scaled, blank_here):
    """Fit the changing coulometer blank during an analysis session."""
    return blank_progression(x0, datenum_scaled) - blank_here


def _get_session_blanks(dbs_group):
    """[group.apply] Calculate blanks per analysis session."""
    x = dbs_group
    blank_here = x[x.blank_good].blank_here
    datenum_here = x[x.blank_good].datenum_analysis
    datenum_mean = datenum_here.mean()
    datenum_std = datenum_here.std()
    datenum_scaled = centre_and_scale(
        datenum_here, x_factor=datenum_std, x_offset=datenum_mean
    )
    blank_prog = least_squares(
        fit_blank_progression, [0, 1, 1, 1, 1], args=[datenum_scaled, blank_here]
    )
    return pd.Series(
        data={
            "blank_mean": blank_here.mean(),
            "blank_median": blank_here.median(),
            "datenum_analysis_mean": datenum_mean,
            "datenum_analysis_std": datenum_std,
            "blank_progression": blank_prog["x"],
        }
    )


def get_session_blanks(dbs, batch_col="dic_cell_id", **kwargs):
    """Calculate blanks per analysis session."""
    if "blank_here" not in dbs:
        dbs.get_sample_blanks(**kwargs)
    if "blank_good" not in dbs:
        dbs["blank_good"] = True
    dbs.sessions = dbs.groupby(by=batch_col).apply(_get_session_blanks)
    return dbs


def blank_correction(
    dbs, blank_col="blank", counts_col="counts", runtime_col="run_time"
):
    """Apply the blank correction to the coulometer counts."""
    return dbs[counts_col] - dbs[runtime_col] * dbs[blank_col]


def get_blank_corrections(
    dbs, blank_col="blank", counts_col="counts", runtime_col="run_time", **kwargs
):
    """Determine and apply the blank correction."""
    if ~isinstance(dbs.sessions, pd.DataFrame):
        dbs.get_session_blanks(**kwargs)
    dbs["datenum_analysis_scaled"] = np.nan
    dbs["blank"] = np.nan
    for session, s in dbs.sessions.iterrows():
        l = dbs[dbs.sessions.index.name] == session
        dbs.loc[l, "datenum_analysis_scaled"] = centre_and_scale(
            dbs.loc[l].datenum_analysis,
            x_factor=s.datenum_analysis_std,
            x_offset=s.datenum_analysis_mean,
        )
        dbs.loc[l, "blank"] = blank_progression(
            s.blank_progression, dbs.loc[l].datenum_analysis_scaled,
        )
    dbs["counts_corrected"] = blank_correction(
        dbs, blank_col=blank_col, counts_col=counts_col, runtime_col=runtime_col
    )
    return dbs


def get_density(dbs):
    """Calculate sample densities in kg/l."""
    if "temperature_analysis_dic" not in dbs:
        dbs["temperature_analysis_dic"] = 25.0
        print("dbs.temperature_analysis_dic not set, so assuming 25 °C.")
    if "salinity" not in dbs:
        dbs["salinity"] = 35.0
        print("dbs.salinity not set, so assuming 35.")
    dbs["density_analysis_dic"] = density.seawater_1atm_MP81(
        temperature=dbs.temperature_analysis_dic, salinity=dbs.salinity
    )
    return dbs


def get_standard_calibrations(dbs, **kwargs):
    """Calculate the separate calibration factor for each CRM."""
    assert "dic_certified" in dbs, "You must provide some dbs.dic_certified values."
    if "counts_corrected" not in dbs:
        dbs.get_blank_corrections(**kwargs)
    if "density_analysis_dic" not in dbs:
        dbs.get_density()
    dbs["k_dic_here"] = (
        dbs.dic_certified * dbs.density_analysis_dic / dbs.counts_corrected
    )
    return dbs


def _get_session_calibrations(dbs_group):
    """[group.apply] Calculate the session-averaged calibration factors."""
    gk = dbs_group[dbs_group.k_dic_good].k_dic_here
    return pd.Series(
        {
            "k_dic_mean": gk.mean(),
            "k_dic_std": gk.std(),
            "k_dic_count": np.sum(~np.isnan(gk)),
        }
    )


def get_session_calibrations(dbs, batch_col="dic_cell_id", **kwargs):
    """Calculate the session-averaged calibration factors."""
    if "k_dic_here" not in dbs:
        dbs.get_standard_calibrations(batch_col=batch_col, **kwargs)
    if "k_dic_good" not in dbs:
        dbs["k_dic_good"] = ~np.isnan(dbs.dic_certified)
    sc = dbs.groupby(by=batch_col).apply(_get_session_calibrations)
    for k, v in sc.iteritems():
        dbs.sessions[k] = v
    dbs["k_dic"] = dbs.sessions.loc[dbs[batch_col]].k_dic_mean.values
    return dbs


def calibrate_dic(dbs, **kwargs):
    """Calibrate all DIC measurements."""
    if "k_dic" not in dbs:
        dbs.get_session_calibrations()
    dbs["dic"] = dbs["counts_corrected"] * dbs["k_dic"] / dbs["density_analysis_dic"]
    dbs["dic_offset"] = dbs.dic - dbs.dic_certified
    return dbs



import pandas as pd


class Dbs(pd.DataFrame):
    logfile = None
    sessions = None

    from .get import (
        get_logfile_index,
        get_sample_blanks,
        get_session_blanks,
        get_blank_corrections,
        get_density,
        get_standard_calibrations,
        get_session_calibrations,
        calibrate_dic,
    )
    from .plot import plot_blanks, plot_session_blanks, plot_k_dic, plot_dic_offset


def concat(objs, logfile=None, **kwargs):
    """Concatenate imported .dbs files."""
    obj = pd.concat(objs, **kwargs)
    obj.reset_index(drop=True, inplace=True)
    obj = Dbs(obj)
    obj.logfile = logfile
    return obj



"""Miscellaneous tools for marine carbonate chemistry and other such things."""

import string, textwrap
import numpy as np
from . import crm, infrared, molar, spectro, vindta
from .vindta import (
    Dbs,
    read_dbs,
    read_logfile,
    get_logfile_index,
    get_sample_blanks,
    blank_progression,
    centre_and_scale,
    de_centre_and_scale,
    get_session_blanks,
    blank_correction,
    get_blank_corrections,
    get_density,
    get_standard_calibrations,
    get_session_calibrations,
    calibrate_dic,
    plot_session_blanks,
    plot_blanks,
    plot_k_dic,
    plot_dic_offset,
    concat,
    poison_correction,
)

__all__ = ["crm", "infrared", "molar", "spectro", "vindta"]
__version__ = "0.0.10"


lcletter = dict(zip(range(1, 27), string.ascii_lowercase))


def sigfig(x, sf):
    """Return `x` to `sf` significant figures."""
    factor = 10.0 ** np.ceil(np.log10(np.abs(x)))
    return factor * np.around(x / factor, decimals=sf)


def say_hello():
    greeting = textwrap.dedent(
        r"""
        k  Miscellaneous tools for
        o  marine carbonate chemistry
        o  and other such things
        l  
        s  Matthew P. Humphreys
        t  https://mvdh.xyz
        o  
        f  Version {}
        """.format(
            __version__
        )
    )
    print(greeting)



import numpy as np


def pH_NIOZ(
    absorbance_578nm, absorbance_434nm, absorbance_730nm, temperature=25, salinity=35
):
    """Calculate pH from NIOZ spectrophotometer absorbances."""
    WL1 = absorbance_578nm
    WL2 = absorbance_434nm
    WL3 = absorbance_730nm
    pH_total = (
        np.log10(
            (((WL1 - WL3) / (WL2 - WL3) - 0.00815 * WL1) - 0.00691)
            / (2.222 - ((WL1 - WL3) / (WL2 - WL3) - 0.00815 * WL1) * 0.1331)
        )
        + 1245.69 / (temperature + 273.15)
        + 3.8275
        + 0.00211 * (35 - salinity)
    )
    return pH_total



import re
import numpy as np, pandas as pd
from matplotlib import dates as mdates
from . import Dbs


def read_logfile(fname, methods="3C standard"):
    """Import a logfile.bak as a DataFrame.
    
    Arguments:
    fname -- the filename (and path) of the .dbs file

    Keywords arguments:
    methods -- list of VINDTA methods considered as real measurements
    """
    if isinstance(methods, str):
        methods = [methods]
    # Compile regexs for reading logfile
    re_method = re.compile(
        r"(" + r"|".join(methods) + r")\.mth run started ".format(methods)
    )
    re_datetime = re.compile(r"started (\d{2})/(\d{2})/(\d{2})  (\d{2}):(\d{2})")
    re_bottle = re.compile(r"(bottle)?\t([^\t]*)\t")
    re_crm = re.compile(r"CRM\t([^\t]*)\t")
    re_increments = re.compile(r"(\d*)\t(\d*)\t(\d*)\t")
    # Import text from logfile
    with open(fname, "r") as f:
        logf = f.read().splitlines()
    # Initialise arrays
    logdf = {
        "line_number": [],
        "datetime_analysis": np.array([], dtype="datetime64"),
        "bottle": [],
        "table": [],
        "counts": [],
        "run_time": [],
        "method": [],
    }
    # Parse file line by line
    for i, line in enumerate(logf):
        if re_method.match(line):
            logdf["line_number"].append(i)
            logdf["method"].append(re_method.findall(line)[0])
            # Get analysis date and time
            ldt = re_datetime.findall(line)[0]
            ldt = np.datetime64(
                "{}-{}-{}T{}:{}".format("20" + ldt[2], ldt[0], ldt[1], ldt[3], ldt[4])
            )
            logdf["datetime_analysis"] = np.append(logdf["datetime_analysis"], ldt)
            # Get sample name
            lbot = 0
            if re_bottle.match(logf[i + 1]):
                lbot = re_bottle.findall(logf[i + 1])[0][1]
            elif re_crm.match(logf[i + 1]):
                lbot = re_crm.findall(logf[i + 1])[0]
            elif logf[i + 1] == "other":
                lbot = "other_{}".format(i + 1)
            assert type(lbot) == str, "Logfile line {}: bottle name not found!".format(
                i + 1
            )
            logdf["bottle"].append(lbot)
            # Get coulometer data
            jdict = {"minutes": [0.0], "counts": [0.0], "increments": [0.0]}
            j = 4
            while re_increments.match(logf[i + j].strip()):
                jinc = re_increments.findall(logf[i + j].strip())[0]
                jdict["minutes"].append(float(jinc[0]))
                jdict["counts"].append(float(jinc[1]))
                jdict["increments"].append(float(jinc[2]))
                j += 1
            jdict = {k: np.array(v) for k, v in jdict.items()}
            logdf["table"].append(jdict)
            logdf["counts"].append(jdict["counts"][-1])
            logdf["run_time"].append(j - 4.0)
    # Convert lists to arrays and put logfile into DataFrame
    logdf = pd.DataFrame({k: np.array(v) for k, v in logdf.items()})
    logdf.set_index("line_number", inplace=True)
    return logdf


# More Python-friendly names for the .dbs columns
dbs_mapper = {
    "run type": "run_type",
    "i.s. temperature": "temperature_insitu",
    "run time": "run_time",
    "CT": "dic_raw",
    "factor CT": "factor_dic",
    "blank": "blank_setting",
    "TCT": "total_carbon_titrated",
    "last CRM CT": "last_crm_dic",
    "cert. CRM CT": "dic_cert",
    "last CRM AT": "last_crm_alkalinity",
    "cert. CRM AT": "alkalinity_cert",
    "batch": "batch_setting",
    "AT": "alkalinity_raw",
    "factor AT": "factor_alkalinity",
    "rms": "rms_alkalinity",
    "calc ID": "calc_id",
    "Titrino": "titrino",
    "sample line": "sample_line",
    "pip vol": "pip_vol",
    "Lat.": "latitude",
    "Long.": "longitude",
    "cell ID": "dic_cell_id",
}


# List of columns to drop from the .dbs by default
dbs_drop = [
    "run_type",
    "temperature_insitu",
    "salinity",
    "dic_raw",
    "factor_dic",
    "blank_setting",
    "total_carbon_titrated",
    "last_crm_dic",
    "dic_cert",
    "last_crm_alkalinity",
    "alkalinity_cert",
    "batch_setting",
    "alkalinity_raw",
    "factor_alkalinity",
    "rms_alkalinity",
    "calc_id",
    "titrino",
    "sample_line",
    "pip_vol",
    "comment",
    "latitude",
    "longitude",
    "date",
    "time",
]


def dbs_datetime(dbsx):
    """[row.apply] Convert date and time from .dbs file into datetime and datenum."""
    dspl = dbsx["date"].split("/")
    datetime_analysis = np.datetime64(
        "-".join(("20" + dspl[2], dspl[0], dspl[1])) + "T" + dbsx["time"]
    )
    return pd.Series(
        {
            "datetime_analysis": datetime_analysis,
            "datenum_analysis": mdates.date2num(datetime_analysis),
        }
    )


def read_dbs(fname, keep_all_cols=False, logfile=None):
    """Import a .dbs file from a VINDTA, rename the columns, and reformat the date/time.
    
    Arguments:
    fname -- the filename (and path) of the .dbs file
    
    Keyword arguments:
    keep_all_cols -- keep all original columns from the .dbs?
    logfile -- the DataFrame containing the logfile.bak
    """
    headers = np.genfromtxt(fname, delimiter="\t", dtype=str, max_rows=1)
    dbs = pd.read_table(fname, header=0, names=headers, usecols=headers).rename(
        columns=dbs_mapper
    )
    dbs["dbs_fname"] = fname
    dbs = Dbs(dbs.assign(**dbs.apply(dbs_datetime, axis=1)))
    if not keep_all_cols:
        dbs.drop(columns=dbs_drop, inplace=True)
    if logfile is not None:
        assert isinstance(logfile, pd.DataFrame), "`logfile` must be a DataFrame."
        dbs.logfile = logfile
    return dbs



import setuptools
import koolstof as ks

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="koolstof",
    version=ks.__version__,
    author="Humphreys, Matthew P.",
    author_email="m.p.humphreys@icloud.com",
    description="Miscellaneous tools for marine carbonate chemistry and other such things",
    url="https://github.com/mvdh7/koolstof",
    packages=setuptools.find_packages(),
    install_requires=["numpy>=1.17", "pandas>=1", "matplotlib>=3.2", "calkulate>=3"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



"""Make figures to assist calibrating and QCing VINDTA datasets."""
import itertools, copy
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.dates as mdates
from . import get, process


markers = itertools.cycle(("o", "^", "s", "v", "D", "<", ">"))
colours = itertools.cycle(
    (
        "xkcd:purple",
        "xkcd:green",
        "xkcd:blue",
        "xkcd:pink",
        "xkcd:brown",
        "xkcd:red",
        "xkcd:teal",
        "xkcd:orange",
    )
)


def increments(dbs, logfile, use_from, ax=None, alpha=0.25, **kwargs):
    """Plot coulometer increments by the minute, focussing on the tails.
    Any additional kwargs are passed to plt.plot().
    """
    if ax is None:
        _, ax = plt.subplots()
    fymax = 1.0
    for i in dbs.logfile_index:
        i_data = logfile.table[i]
        i_blank = i_data["minutes"] >= use_from
        ax.plot(
            i_data["minutes"][i_blank],
            i_data["increments"][i_blank],
            c="xkcd:strawberry",
            alpha=alpha,
        )
        fymax = np.max([fymax, np.max(i_data["increments"][-3:])])
        not_i_blank = i_data["minutes"] <= use_from
        ax.plot(
            i_data["minutes"][not_i_blank],
            i_data["increments"][not_i_blank],
            c="xkcd:navy",
            alpha=alpha,
        )
    ax.set_xlim([0, dbs["run time"].max()])
    ax.set_ylim([0, fymax * 1.2])
    ax.set_xlabel("Run time / minutes")
    ax.set_ylabel("Increments / per minute")
    return ax


def plot_session_blanks(
    dbs,
    session,
    ax=None,
    c="xkcd:navy",
    marker="o",
    figure_path=None,
    figure_format="png",
):
    """Draw sample blanks and their fit for one analysis session."""
    # Prepare to draw the figure
    s = dbs.sessions.loc[session]
    l = dbs["dic_cell_id"] == session
    if ax is None:
        fig, ax = plt.subplots(dpi=300)
    # Create and draw fitted line
    fx = np.linspace(
        dbs[l].datenum_analysis_scaled.min(), dbs[l].datenum_analysis_scaled.max(), 500
    )
    fy = get.blank_progression(s.blank_progression, fx)
    fx = mdates.num2date(
        get.de_centre_and_scale(fx, s.datenum_analysis_std, s.datenum_analysis_mean)
    )
    ax.plot(fx, fy, c=c, label="Best fit")
    # Draw the rest of the figure
    dbs[l & dbs.blank_good].plot.scatter(
        "datetime_analysis",
        "blank_here",
        ax=ax,
        c=c,
        marker=marker,
        label="Samples used",
    )
    dbs[l & ~dbs.blank_good].plot.scatter(
        "datetime_analysis",
        "blank_here",
        ax=ax,
        c="none",
        edgecolor=c,
        marker=marker,
        label="Ignored",
    )
    y_max = np.max([dbs[l & dbs.blank_good].blank_here.max(), np.max(fy)]) * 1.05
    off_x = dbs[l & (dbs.blank_here > y_max)].datetime_analysis.values
    ax.scatter(
        off_x,
        np.full(np.size(off_x), y_max * 0.99999),
        c="none",
        edgecolor=c,
        marker="^",
        label="Off scale (ignored)",
        clip_on=False,
    )
    ax.set_ylim([0, y_max])
    ax.legend(edgecolor="k")
    ax.xaxis.set_major_locator(mdates.HourLocator())
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H"))
    ax.set_xlabel("Time of day of analysis")
    ax.set_ylabel(r"Coulometer blank / count$\cdot$minute$^{-1}$")
    ax.set_title(session)
    ax.grid(alpha=0.2)
    plt.tight_layout()
    if figure_path is not None:
        plt.savefig("{}/{}.{}".format(figure_path, str(session), figure_format))
    return fig, ax


def plot_blanks(dbs, **kwargs):
    """Draw sample blanks and their fit for all analysis sessions."""
    for session in dbs.sessions.index:
        fig, _ = plot_session_blanks(dbs, session, **kwargs)
        plt.close(fig)


def blanks(dbs, dic_sessions, ax=None, title=None, alpha=0.5, **kwargs):
    """Plot sample-by-sample blank values.
    Additional kwargs are passed to plt.scatter.
    """
    if ax is None:
        _, ax = plt.subplots()
    if "blank_good" not in dbs:
        dbs["blank_good"] = True
    dbs[~dbs.blank_good].plot.scatter(
        "analysis_datetime",
        "blank_here",
        ax=ax,
        c="xkcd:strawberry",
        alpha=alpha,
        **kwargs
    )
    dbs[dbs.blank_good].plot.scatter(
        "analysis_datetime", "blank_here", ax=ax, c="xkcd:navy", alpha=alpha, **kwargs
    )
    sessions_here = dbs["cell ID"].unique()
    for session in sessions_here:
        sl = dbs["cell ID"] == session
        sx_sc = process.centre_and_scale(
            np.linspace(
                np.min(dbs[sl]["analysis_datenum"]),
                np.max(dbs[sl]["analysis_datenum"]),
                num=100,
            ),
            x_factor=dic_sessions.loc[session].analysis_datenum_std,
            x_offset=dic_sessions.loc[session].analysis_datenum_mean,
        )
        sx = mdates.num2date(
            process.de_centre_and_scale(
                sx_sc,
                dic_sessions.loc[session].analysis_datenum_std,
                dic_sessions.loc[session].analysis_datenum_mean,
            )
        )
        ax.plot(
            sx,
            process.blank_progression(
                dic_sessions.loc[session].blank_progression, sx_sc
            ),
            c="xkcd:navy",
        )
    ax.set_xlim(
        [
            dbs.analysis_datetime.min() - np.timedelta64(30, "m"),
            dbs.analysis_datetime.max() + np.timedelta64(30, "m"),
        ]
    )
    ax.xaxis.set_major_locator(mdates.HourLocator())
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H"))
    ax.set_title(title)
    ax.set_xlabel("Time of day")
    ax.set_ylim([0, np.max(dbs[dbs.blank_good].blank_here) * 1.05])
    ax.set_ylabel("Sample blank / count per minute")
    return ax


def plot_k_dic(
    dbs, ax=None, batch_col="dic_cell_id", figure_path=None, figure_format="png"
):
    """Plot DIC calibration factors through time."""
    marker = copy.deepcopy(markers)
    colour = copy.deepcopy(colours)
    if ax is None:
        fig, ax = plt.subplots(dpi=300)
    for session, s in dbs.sessions.iterrows():
        m = next(marker)
        c = next(colour)
        l = dbs[batch_col] == session
        l_good = l & dbs.k_dic_good
        if l_good.any():
            dbs[l_good].plot.scatter(
                "datetime_analysis",
                "k_dic_here",
                ax=ax,
                c=c,
                marker=m,
                alpha=0.7,
                label=session,
            )
        l_bad = l & ~dbs.k_dic_good & ~np.isnan(dbs.dic_certified)
        if l_bad.any():
            dbs[l_bad].plot.scatter(
                "datetime_analysis",
                "k_dic_here",
                ax=ax,
                c="none",
                edgecolor=c,
                marker=m,
            )
        sl = dbs[batch_col] == session
        sx = np.array(
            [
                dbs.loc[sl, "datetime_analysis"].min(),
                dbs.loc[sl, "datetime_analysis"].max(),
            ]
        )
        sy = np.full_like(sx, s.k_dic_mean)
        ax.plot(sx, sy, c=c)
    ax.legend(edgecolor="k", bbox_to_anchor=(1, 1))
    ax.set_xlabel("Analysis date and time")
    ax.set_ylabel(r"DIC calibration factor / μmol$\cdot$count$^{-1}$")
    fac_mean = dbs.k_dic.mean()
    fac_maxdiff = (dbs.k_dic - fac_mean).abs().max()
    if fac_maxdiff > 0:
        ax.set_ylim(np.array([-1, 1]) * fac_maxdiff * 1.1 + fac_mean)
    # ax.set_xlim(
    #     [
    #         dbs.datetime_analysis.min() - np.timedelta64(30, "m"),
    #         dbs.datetime_analysis.max() + np.timedelta64(30, "m"),
    #     ]
    # )
    ax.grid(alpha=0.2)
    plt.tight_layout()
    if figure_path is not None:
        plt.savefig("{}/k_dic.{}".format(figure_path, figure_format))
    return fig, ax


def plot_dic_offset(
    dbs, ax=None, batch_col="dic_cell_id", figure_path=None, figure_format="png"
):
    """Plot measured minus certified DIC values through time."""
    marker = copy.deepcopy(markers)
    colour = copy.deepcopy(colours)
    if ax is None:
        fig, ax = plt.subplots(dpi=300)
    for session in dbs.sessions.index:
        m = next(marker)
        c = next(colour)
        l = dbs[batch_col] == session
        l_good = l & dbs.k_dic_good
        if l_good.any():
            dbs[l_good].plot.scatter(
                "datetime_analysis",
                "dic_offset",
                ax=ax,
                c=c,
                marker=m,
                alpha=0.7,
                label=session,
            )
        l_bad = l & ~dbs.k_dic_good & ~np.isnan(dbs.dic_offset)
        if l_bad.any():
            dbs[l_bad].plot.scatter(
                "datetime_analysis",
                "dic_offset",
                ax=ax,
                c="none",
                edgecolor=c,
                marker=m,
            )
    ax.legend(edgecolor="k", bbox_to_anchor=(1, 1))
    ax.set_xlabel("Analysis date and time")
    ax.set_ylabel(r"DIC (calibrated $-$ certified) / μmol$\cdot$kg$^{-1}$")
    ax.set_ylim(np.array([-1, 1]) * dbs[dbs.k_dic_good].dic_offset.abs().max() * 1.1)
    # ax.set_xlim(
    #     [
    #         dbs.datetime_analysis.min() - np.timedelta64(30, "m"),
    #         dbs.datetime_analysis.max() + np.timedelta64(30, "m"),
    #     ]
    # )
    ax.grid(alpha=0.2)
    ax.axhline(0, c="k", linewidth=0.8)
    plt.tight_layout()
    if figure_path is not None:
        plt.savefig("{}/dic_offset.{}".format(figure_path, figure_format))
    return fig, ax


def dic_calibration_factors(dbs, dic_sessions, ax=None, save_as=None, **kwargs):
    """Plot changing DIC calibration factors through time."""
    if ax is None:
        _, ax = plt.subplots()
    if dbs.dic_calibration_good.any():
        dbs[dbs.dic_calibration_good].plot.scatter(
            "analysis_datetime",
            "dic_calibration_factor",
            ax=ax,
            c="xkcd:navy",
            alpha=0.5,
            **kwargs
        )
    if (~dbs.dic_calibration_good & dbs.crm).any():
        dbs[~dbs.dic_calibration_good].plot.scatter(
            "analysis_datetime",
            "dic_calibration_factor",
            ax=ax,
            c="xkcd:strawberry",
            alpha=0.5,
            **kwargs
        )
    for session in dic_sessions.index:
        if session in dbs["cell ID"].values:
            sl = dbs["cell ID"] == session
            sx = np.array(
                [
                    dbs.loc[sl, "analysis_datetime"].min(),
                    dbs.loc[sl, "analysis_datetime"].max(),
                ]
            )
            sy = np.full_like(sx, dic_sessions.loc[session].dic_calibration_mean)
            ax.plot(sx, sy, c="xkcd:strawberry", zorder=100, alpha=1)
    ax.set_ylabel("DIC calibration factor / μmol/count")
    fac_mean = dbs.dic_calibration_factor.mean()
    fac_maxdiff = (dbs.dic_calibration_factor - fac_mean).abs().max()
    if fac_maxdiff > 0:
        ax.set_ylim(np.array([-1, 1]) * fac_maxdiff * 1.1 + fac_mean)
    ax.set_xlim(
        [
            dbs.analysis_datetime.min() - np.timedelta64(30, "m"),
            dbs.analysis_datetime.max() + np.timedelta64(30, "m"),
        ]
    )
    ax.grid(alpha=0.4)
    return ax



"""Certified reference material (CRM) data."""

import numpy as np


class ReferenceMaterial:
    """A generic reference material."""

    def __init__(self, **kwargs):
        assert all(
            [
                key
                in [
                    "batch",
                    "salinity",
                    "dissolved_inorganic_carbon",
                    "dissolved_inorganic_carbon_std",
                    "total_alkalinity",
                    "total_alkalinity_std",
                    "phosphate",
                    "silicate",
                    "nitrite",
                    "nitrate",
                ]
                for key in kwargs
            ]
        )
        self.__dict__.update((k, v) for k, v in kwargs.items())


dickson_certified_values = {
    171: ReferenceMaterial(
        batch=171,
        salinity=33.434,
        dissolved_inorganic_carbon=2029.19,
        dissolved_inorganic_carbon_std=0.87,
        total_alkalinity=2217.40,
        total_alkalinity_std=0.63,
        phosphate=0.43,
        silicate=2.3,
        nitrite=0.00,
        nitrate=3.1,
    ),
    186: ReferenceMaterial(
        batch=186,
        salinity=33.525,
        dissolved_inorganic_carbon=2012.59,
        dissolved_inorganic_carbon_std=0.80,
        total_alkalinity=2212.00,
        total_alkalinity_std=0.53,
        phosphate=0.42,
        silicate=3.3,
        nitrite=0.01,
        nitrate=2.8,
    ),
}


def dickson(crm_batches, fields):
    """Dickson seawater certified reference material."""
    if isinstance(fields, str):
        if fields == "all":
            fields = [
                "batch",
                "salinity",
                "dissolved_inorganic_carbon",
                "dissolved_inorganic_carbon_std",
                "total_alkalinity",
                "total_alkalinity_std",
                "phosphate",
                "silicate",
                "nitrite",
                "nitrate",
            ]
        else:
            fields = [fields]
    return {
        field: np.array(
            [
                dickson_certified_values[crm_batch].__dict__[field]
                if crm_batch in dickson_certified_values
                else np.nan
                for crm_batch in crm_batches
            ]
        )
        for field in fields
    }


def dic_calibration_factor(
    data=None,
    dic_certified="dic_certified",
    density_analysis="density_analysis",
    counts_corrected="counts_corrected",
):
    """Calculate volume-cancelling DIC calibration factor."""
    if data is not None:
        dic_certified = data[dic_certified]
        density_analysis = data[density_analysis]
        counts_corrected = data[counts_corrected]
    return dic_certified * density_analysis / counts_corrected


def calibrate_dic(
    data=None,
    calibration_factor="dic_calibration_factor",
    counts_corrected="counts_corrected",
    density_analysis="density_analysis",
):
    """Apply volume-cancelling DIC calibration factor."""
    if data is not None:
        calibration_factor = data[calibration_factor]
        counts_corrected = data[counts_corrected]
        density_analysis = data[density_analysis]
    return calibration_factor * counts_corrected / density_analysis
